from django.db import models
from django.db.models import TextChoices

from canvas_sdk.v1.data.base import CommittableModelManager, ValueSetLookupQuerySet


class Status(TextChoices):
    """Medication status."""

    ACTIVE = "active", "active"
    INACTIVE = "inactive", "inactive"


class MedicationQuerySet(ValueSetLookupQuerySet):
    """MedicationQuerySet."""

    pass


class Medication(models.Model):
    """Medication."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_medication_001"

    objects = CommittableModelManager.from_queryset(MedicationQuerySet)()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="medications", null=True
    )
    deleted = models.BooleanField()
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    status = models.CharField(choices=Status.choices)
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    quantity_qualifier_description = models.CharField()
    clinical_quantity_description = models.CharField()
    potency_unit_code = models.CharField()
    national_drug_code = models.CharField()
    erx_quantity = models.CharField()


class MedicationCoding(models.Model):
    """MedicationCoding."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_medicationcoding_001"

    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    medication = models.ForeignKey(
        Medication, on_delete=models.DO_NOTHING, related_name="codings", null=True
    )



from canvas_sdk.effects import Effect
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol
from logger import log


class Forbidden(BaseProtocol):
    """You should put a helpful description of this protocol's behavior here."""

    # Name the event type you wish to run in response to
    RESPONDS_TO = EventType.Name(EventType.UNKNOWN)

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        from test_implicit_imports_plugin.utils.base import OtherClass

        OtherClass()

        return []


class Allowed(BaseProtocol):
    """You should put a helpful description of this protocol's behavior here."""

    RESPONDS_TO = EventType.Name(EventType.UNKNOWN)

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        from test_implicit_imports_plugin.templates import Template

        log.info(Template().render())

        return []



from abc import abstractmethod
from enum import StrEnum

from canvas_sdk.effects import Effect
from canvas_sdk.effects.show_button import ShowButtonEffect
from canvas_sdk.events import EventType
from canvas_sdk.handlers.base import BaseHandler


class ActionButton(BaseHandler):
    """Base class for action buttons."""

    RESPONDS_TO = [
        EventType.Name(EventType.SHOW_NOTE_HEADER_BUTTON),
        EventType.Name(EventType.SHOW_NOTE_FOOTER_BUTTON),
        EventType.Name(EventType.ACTION_BUTTON_CLICKED),
    ]

    class ButtonLocation(StrEnum):
        NOTE_HEADER = "note_header"
        NOTE_FOOTER = "note_footer"

    BUTTON_TITLE: str = ""
    BUTTON_KEY: str = ""
    BUTTON_LOCATION: ButtonLocation | None = None

    @abstractmethod
    def handle(self) -> list[Effect]:
        """Method to handle button click."""
        raise NotImplementedError("Implement to handle button click")

    def visible(self) -> bool:
        """Method to determine button visibility."""
        return True

    def compute(self) -> list[Effect]:
        """Method to compute the effects."""
        if self.BUTTON_LOCATION is None:
            return []

        if self.event.type in (
            EventType.SHOW_NOTE_HEADER_BUTTON,
            EventType.SHOW_NOTE_FOOTER_BUTTON,
        ):
            if (
                self.event.context["location"].lower() == self.BUTTON_LOCATION.value
                and self.visible()
            ):
                return [ShowButtonEffect(key=self.BUTTON_KEY, title=self.BUTTON_TITLE).apply()]
            else:
                return []
        elif (
            self.event.type == EventType.ACTION_BUTTON_CLICKED
            and self.event.context["key"] == self.BUTTON_KEY
        ):
            return self.handle()

        return []



import logging
import shutil
from collections.abc import Generator
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from canvas_generated.messages.effects_pb2 import EffectType
from canvas_generated.messages.plugins_pb2 import ReloadPluginsRequest
from canvas_sdk.events import Event, EventRequest, EventType
from plugin_runner.plugin_runner import (
    EVENT_HANDLER_MAP,
    LOADED_PLUGINS,
    PluginRunner,
    load_or_reload_plugin,
    load_plugins,
)


@pytest.fixture
def install_test_plugin(request: pytest.FixtureRequest) -> Generator[Path, None, None]:
    """Copies a specified plugin from the fixtures directory to the data directory
    and removes it after the test.

    Parameters:
    - request.param: The name of the plugin package to copy.

    Yields:
    - Path to the copied plugin directory.
    """
    # Define base directories
    base_dir = Path("./plugin_runner/tests")
    fixture_plugin_dir = base_dir / "fixtures" / "plugins"
    data_plugin_dir = base_dir / "data" / "plugins"

    # The plugin name should be passed as a parameter to the fixture
    plugin_name = request.param  # Expected to be a str
    src_plugin_path = fixture_plugin_dir / plugin_name
    dest_plugin_path = data_plugin_dir / plugin_name

    # Ensure the data plugin directory exists
    data_plugin_dir.mkdir(parents=True, exist_ok=True)

    # Copy the specific plugin from fixtures to data
    try:
        shutil.copytree(src_plugin_path, dest_plugin_path)
        yield dest_plugin_path  # Provide the path to the test
    finally:
        # Cleanup: remove data/plugins directory after the test
        if dest_plugin_path.exists():
            shutil.rmtree(dest_plugin_path)


@pytest.fixture
def load_test_plugins() -> Generator[None, None, None]:
    """Manages the lifecycle of test plugins by loading and unloading them."""
    try:
        load_plugins()
        yield
    finally:
        LOADED_PLUGINS.clear()
        EVENT_HANDLER_MAP.clear()


@pytest.fixture
def plugin_runner() -> PluginRunner:
    """Fixture to initialize PluginRunner with mocks."""
    runner = PluginRunner()
    runner.statsd_client = MagicMock()
    return runner


@pytest.mark.parametrize("install_test_plugin", ["example_plugin"], indirect=True)
def test_load_plugins_with_valid_plugin(install_test_plugin: Path, load_test_plugins: None) -> None:
    """Test loading plugins with a valid plugin."""
    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS
    assert (
        LOADED_PLUGINS["example_plugin:example_plugin.protocols.my_protocol:Protocol"]["active"]
        is True
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("install_test_plugin", ["test_module_imports_plugin"], indirect=True)
async def test_load_plugins_with_plugin_that_imports_other_modules_within_plugin_package(
    install_test_plugin: Path, plugin_runner: PluginRunner, load_test_plugins: None
) -> None:
    """Test loading plugins with a valid plugin that imports other modules within the current plugin package."""
    assert (
        "test_module_imports_plugin:test_module_imports_plugin.protocols.my_protocol:Protocol"
        in LOADED_PLUGINS
    )
    assert (
        LOADED_PLUGINS[
            "test_module_imports_plugin:test_module_imports_plugin.protocols.my_protocol:Protocol"
        ]["active"]
        is True
    )

    result = [
        response
        async for response in plugin_runner.HandleEvent(EventRequest(type=EventType.UNKNOWN), None)
    ]

    assert len(result) == 1
    assert result[0].success is True
    assert len(result[0].effects) == 1
    assert result[0].effects[0].type == EffectType.LOG
    assert result[0].effects[0].payload == "Successfully imported!"


@pytest.mark.parametrize(
    "install_test_plugin",
    [
        "test_module_imports_outside_plugin_v1",
        "test_module_imports_outside_plugin_v2",
        "test_module_imports_outside_plugin_v3",
    ],
    indirect=True,
)
def test_load_plugins_with_plugin_that_imports_other_modules_outside_plugin_package(
    install_test_plugin: Path, caplog: pytest.LogCaptureFixture
) -> None:
    """Test loading plugins with an invalid plugin that imports other modules outside the current plugin package."""
    with caplog.at_level(logging.ERROR):
        load_or_reload_plugin(install_test_plugin)

    assert any(
        "Error importing module" in record.message for record in caplog.records
    ), "log.error() was not called with the expected message."


@pytest.mark.parametrize(
    "install_test_plugin",
    [
        "test_module_forbidden_imports_plugin",
    ],
    indirect=True,
)
def test_load_plugins_with_plugin_that_imports_forbidden_modules(
    install_test_plugin: Path, caplog: pytest.LogCaptureFixture
) -> None:
    """Test loading plugins with an invalid plugin that imports forbidden modules."""
    with caplog.at_level(logging.ERROR):
        load_or_reload_plugin(install_test_plugin)

    assert any(
        "Error importing module" in record.message for record in caplog.records
    ), "log.error() was not called with the expected message."


@pytest.mark.parametrize(
    "install_test_plugin",
    [
        "test_module_forbidden_imports_runtime_plugin",
    ],
    indirect=True,
)
def test_load_plugins_with_plugin_that_imports_forbidden_modules_at_runtime(
    install_test_plugin: Path,
) -> None:
    """Test loading plugins with an invalid plugin that imports forbidden modules at runtime."""
    with pytest.raises(ImportError, match="is not an allowed import."):
        load_or_reload_plugin(install_test_plugin)
        class_handler = LOADED_PLUGINS[
            "test_module_forbidden_imports_runtime_plugin:test_module_forbidden_imports_runtime_plugin.protocols.my_protocol:Protocol"
        ]["class"]
        class_handler(Event(EventRequest(type=EventType.UNKNOWN))).compute()


@pytest.mark.parametrize(
    "install_test_plugin",
    [
        "test_implicit_imports_plugin",
    ],
    indirect=True,
)
def test_plugin_that_implicitly_imports_allowed_modules(
    install_test_plugin: Path, caplog: pytest.LogCaptureFixture
) -> None:
    """Test loading plugins with a plugin that implicitly imports allowed modules."""
    with caplog.at_level(logging.INFO):
        load_or_reload_plugin(install_test_plugin)
        class_handler = LOADED_PLUGINS[
            "test_implicit_imports_plugin:test_implicit_imports_plugin.protocols.my_protocol:Allowed"
        ]["class"]
        class_handler(Event(EventRequest(type=EventType.UNKNOWN))).compute()

    assert any(
        "Hello, World!" in record.message for record in caplog.records
    ), "log.info() with Template.render() was not called."


@pytest.mark.parametrize(
    "install_test_plugin",
    [
        "test_implicit_imports_plugin",
    ],
    indirect=True,
)
def test_plugin_that_implicitly_imports_forbidden_modules(
    install_test_plugin: Path, caplog: pytest.LogCaptureFixture
) -> None:
    """Test loading plugins with an invalid plugin that implicitly imports forbidden modules."""
    with (
        caplog.at_level(logging.INFO),
        pytest.raises(ImportError, match="'os' is not an allowed import."),
    ):
        load_or_reload_plugin(install_test_plugin)
        class_handler = LOADED_PLUGINS[
            "test_implicit_imports_plugin:test_implicit_imports_plugin.protocols.my_protocol:Forbidden"
        ]["class"]
        class_handler(Event(EventRequest(type=EventType.UNKNOWN))).compute()

    assert (
        any("os list dir" in record.message for record in caplog.records) is False
    ), "log.info() with os.listdir() was called."


@pytest.mark.parametrize("install_test_plugin", ["example_plugin"], indirect=True)
def test_reload_plugin(install_test_plugin: Path, load_test_plugins: None) -> None:
    """Test reloading a plugin."""
    load_plugins()

    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS
    assert (
        LOADED_PLUGINS["example_plugin:example_plugin.protocols.my_protocol:Protocol"]["active"]
        is True
    )


@pytest.mark.parametrize("install_test_plugin", ["example_plugin"], indirect=True)
def test_remove_plugin_should_be_removed_from_loaded_plugins(
    install_test_plugin: Path, load_test_plugins: None
) -> None:
    """Test removing a plugin."""
    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS
    shutil.rmtree(install_test_plugin)
    load_plugins()
    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" not in LOADED_PLUGINS


@pytest.mark.parametrize("install_test_plugin", ["example_plugin"], indirect=True)
@pytest.mark.parametrize("load_test_plugins", [None], indirect=True)
def test_load_plugins_should_refresh_event_protocol_map(
    load_test_plugins: None, install_test_plugin: Path
) -> None:
    """Test that the event protocol map is refreshed when loading plugins."""
    assert EVENT_HANDLER_MAP == {}
    load_plugins()
    assert EventType.Name(EventType.UNKNOWN) in EVENT_HANDLER_MAP
    assert EVENT_HANDLER_MAP[EventType.Name(EventType.UNKNOWN)] == [
        "example_plugin:example_plugin.protocols.my_protocol:Protocol"
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("install_test_plugin", ["example_plugin"], indirect=True)
async def test_handle_plugin_event_returns_expected_result(
    install_test_plugin: Path, plugin_runner: PluginRunner, load_test_plugins: None
) -> None:
    """Test that HandleEvent successfully calls the relevant plugins and returns the expected result."""
    event = EventRequest(type=EventType.UNKNOWN)

    result = []
    async for response in plugin_runner.HandleEvent(event, None):
        result.append(response)

    assert len(result) == 1
    assert result[0].success is True
    assert len(result[0].effects) == 1
    assert result[0].effects[0].type == EffectType.LOG
    assert result[0].effects[0].payload == "Hello, world!"


@pytest.mark.asyncio
@pytest.mark.parametrize("install_test_plugin", ["example_plugin"], indirect=True)
async def test_reload_plugins_event_handler_successfully_publishes_message(
    install_test_plugin: Path, plugin_runner: PluginRunner
) -> None:
    """Test ReloadPlugins Event handler successfully publishes a message with restart action."""
    with patch("plugin_runner.plugin_runner.publish_message", MagicMock()) as mock_publish_message:
        request = ReloadPluginsRequest()

        result = []
        async for response in plugin_runner.ReloadPlugins(request, None):
            result.append(response)

        mock_publish_message.assert_called_once_with({"action": "restart"})

    assert len(result) == 1
    assert result[0].success is True


@pytest.mark.asyncio
@pytest.mark.parametrize("install_test_plugin", ["test_module_imports_plugin"], indirect=True)
async def test_changes_to_plugin_modules_should_be_reflected_after_reload(
    install_test_plugin: Path, load_test_plugins: None, plugin_runner: PluginRunner
) -> None:
    """Test that changes to plugin modules are reflected after reloading the plugin."""
    event = EventRequest(type=EventType.UNKNOWN)

    result = []
    async for response in plugin_runner.HandleEvent(event, None):
        result.append(response)

    assert len(result) == 1
    assert result[0].success is True
    assert len(result[0].effects) == 1
    assert result[0].effects[0].type == EffectType.LOG
    assert result[0].effects[0].payload == "Successfully imported!"

    NEW_CODE = """
def import_me() -> str:
    return "Successfully changed!"
"""
    file_path = install_test_plugin / "other_module" / "base.py"
    file_path.write_text(NEW_CODE, encoding="utf-8")

    # Reload the plugin
    load_plugins()

    result = []
    async for response in plugin_runner.HandleEvent(event, None):
        result.append(response)

    assert len(result) == 1
    assert result[0].success is True
    assert len(result[0].effects) == 1
    assert result[0].effects[0].type == EffectType.LOG
    assert result[0].effects[0].payload == "Successfully changed!"



from time import time
from typing import Any


def get_duration_ms(start_time: float) -> int:
    """Get the duration in milliseconds since the given start time."""
    return int((time() - start_time) * 1000)


LINE_PROTOCOL_TRANSLATION = str.maketrans(
    {
        ",": r"\,",
        "=": r"\=",
        " ": r"\ ",
        ":": r"__",
    }
)


def tags_to_line_protocol(tags: dict[str, Any]) -> str:
    """Generate a tags string compatible with the InfluxDB line protocol.

    See: https://docs.influxdata.com/influxdb/v1.1/write_protocols/line_protocol_tutorial/
    """
    return ",".join(
        f"{tag_name}={str(tag_value).translate(LINE_PROTOCOL_TRANSLATION)}"
        for tag_name, tag_value in tags.items()
    )



import os
from typing import Any
from urllib import parse


def get_database_dict_from_url() -> dict[str, Any]:
    """Retrieves the database URL for the data module connection formatted for Django settings."""
    parsed_url = parse.urlparse(os.getenv("DATABASE_URL"))
    db_name = parsed_url.path[1:]
    return {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": db_name,
        "USER": os.getenv("CANVAS_SDK_DATABASE_ROLE"),
        "PASSWORD": os.getenv("CANVAS_SDK_DATABASE_ROLE_PASSWORD"),
        "HOST": parsed_url.hostname,
        "PORT": parsed_url.port,
    }



from ..value_set import ValueSet


class StatinAllergen(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of an allergy to statin medications.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that define an allergy to statin medications.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS347v5
    """

    VALUE_SET_NAME = "Statin Allergen"
    OID = "2.16.840.1.113762.1.4.1110.42"
    DEFINITION_VERSION = "20210220"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    RXNORM = {
        "6472",  # lovastatin
        "36567",  # simvastatin
        "41127",  # fluvastatin
        "42463",  # pravastatin
        "83367",  # atorvastatin
        "301542",  # rosuvastatin
        "861634",  # pitavastatin
    }
    SNOMEDCT = {
        "372912004",  # Substance with 3-hydroxy-3-methylglutaryl-coenzyme A reductase inhibitor mechanism of action (substance)
    }



from canvas_cli.apps.plugin.plugin import (
    disable,
    enable,
    init,
    install,
    list,
    uninstall,
    validate_manifest,
)

__all__ = ("uninstall", "enable", "disable", "init", "validate_manifest", "install", "list")



import pytest

from canvas_cli.utils.validators import validate_manifest_file


@pytest.fixture
def protocol_manifest_example() -> dict:
    """Return a valid protocol manifest example."""
    return {
        "sdk_version": "0.3.1",
        "plugin_version": "1.0.1",
        "name": "Prompt to prescribe when assessing condition",
        "description": "To assist in ....",
        "components": {
            "protocols": [
                {
                    "class": "prompt_to_prescribe.protocols.prompt_when_assessing.PromptWhenAssessing",
                    "description": "probably the same as the plugin's description",
                    "data_access": {
                        "event": "",
                        "read": ["conditions"],
                        "write": ["commands"],
                    },
                }
            ]
        },
        "tags": {"patient_sourcing_and_intake": ["symptom_triage"]},
        "references": [],
        "license": "",
        "diagram": False,
        "readme": "README.MD",
    }


def test_manifest_file_schema(protocol_manifest_example: dict) -> None:
    """Test that no exception raised when a valid manifest file is validated."""
    validate_manifest_file(protocol_manifest_example)



from enum import Enum

from pydantic import conint, constr

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class VitalsCommand(BaseCommand):
    """A class for managing a Vitals command within a specific note."""

    class Meta:
        key = "vitals"
        commit_required_fields = ()

    class BodyTemperatureSite(Enum):
        AXILLARY = 0
        ORAL = 1
        RECTAL = 2
        TEMPORAL = 3
        TYMPANIC = 4

    class BloodPressureSite(Enum):
        SITTING_RIGHT_UPPER = 0
        SITTING_LEFT_UPPER = 1
        SITTING_RIGHT_LOWER = 2
        SITTING_LEFT_LOWER = 3
        STANDING_RIGHT_UPPER = 4
        STANDING_LEFT_UPPER = 5
        STANDING_RIGHT_LOWER = 6
        STANDING_LEFT_LOWER = 7
        SUPINE_RIGHT_UPPER = 8
        SUPINE_LEFT_UPPER = 9
        SUPINE_RIGHT_LOWER = 10
        SUPINE_LEFT_LOWER = 11

    class PulseRhythm(Enum):
        REGULAR = 0
        IRREGULARLY_IRREGULAR = 1
        REGULARLY_IRREGULAR = 2

    height: conint(ge=10, le=108) | None = None  # type: ignore[valid-type]
    weight_lbs: conint(ge=1, le=1500) | None = None  # type: ignore[valid-type]
    weight_oz: int | None = None
    waist_circumference: conint(ge=20, le=200) | None = None  # type: ignore[valid-type]
    body_temperature: conint(ge=85, le=107) | None = None  # type: ignore[valid-type]
    body_temperature_site: BodyTemperatureSite | None = None
    blood_pressure_systole: conint(ge=30, le=305) | None = None  # type: ignore[valid-type]
    blood_pressure_diastole: conint(ge=20, le=180) | None = None  # type: ignore[valid-type]
    blood_pressure_position_and_site: BloodPressureSite | None = None
    pulse: conint(ge=30, le=250) | None = None  # type: ignore[valid-type]
    pulse_rhythm: PulseRhythm | None = None
    respiration_rate: conint(ge=6, le=60) | None = None  # type: ignore[valid-type]
    oxygen_saturation: conint(ge=60, le=100) | None = None  # type: ignore[valid-type]
    note: constr(max_length=150) | None = None  # type: ignore[valid-type]

    @property
    def values(self) -> dict:
        """The Vitals command's field values."""
        return {
            "height": self.height,
            "weight_lbs": self.weight_lbs,
            "weight_oz": self.weight_oz,
            "waist_circumference": self.waist_circumference,
            "body_temperature": self.body_temperature,
            "body_temperature_site": (
                self.body_temperature_site.value if self.body_temperature_site else None
            ),
            "blood_pressure_systole": self.blood_pressure_systole,
            "blood_pressure_diastole": self.blood_pressure_diastole,
            "blood_pressure_position_and_site": (
                self.blood_pressure_position_and_site.value
                if self.blood_pressure_position_and_site
                else None
            ),
            "pulse": self.pulse,
            "pulse_rhythm": self.pulse_rhythm.value if self.pulse_rhythm else None,
            "respiration_rate": self.respiration_rate,
            "oxygen_saturation": self.oxygen_saturation,
            "note": self.note,
        }



from django.db import models

from canvas_sdk.v1.data.base import CommittableModelManager, ValueSetLookupQuerySet


class AllergyIntolerance(models.Model):
    """AllergyIntolerance."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_allergyintolerance_001"

    objects = CommittableModelManager().from_queryset(ValueSetLookupQuerySet)()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    patient = models.ForeignKey(
        "v1.Patient",
        on_delete=models.DO_NOTHING,
        related_name="allergy_intolerances",
        null=True,
    )
    note_id = models.BigIntegerField()
    allergy_intolerance_type = models.CharField()
    category = models.IntegerField()
    status = models.CharField()
    severity = models.CharField()
    onset_date = models.DateField()
    onset_date_original_input = models.CharField()
    last_occurrence = models.DateField()
    last_occurrence_original_input = models.CharField()
    recorded_date = models.DateTimeField()
    narrative = models.CharField()


class AllergyIntoleranceCoding(models.Model):
    """AllergyIntoleranceCoding."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_allergyintolerancecoding_001"

    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    allergy_intolerance = models.ForeignKey(
        AllergyIntolerance,
        on_delete=models.DO_NOTHING,
        related_name="codings",
        null=True,
    )



import os
import shutil
from collections.abc import Callable, Generator
from datetime import datetime
from pathlib import Path
from typing import Any, cast
from unittest.mock import MagicMock, patch
from urllib.parse import urlparse

import pytest
from django.core.exceptions import ImproperlyConfigured
from typer.testing import CliRunner

import settings
from canvas_cli.apps.auth.utils import CONFIG_PATH

from .main import app

runner = CliRunner()


@pytest.fixture(scope="session")
def plugin_name() -> str:
    """The plugin name to be used for the canvas cli test."""
    return f"cli-{datetime.now().timestamp()}".replace(".", "")


@pytest.fixture(scope="session")
def create_or_update_config_auth_file_for_testing(plugin_name: str) -> Generator[None, None, None]:
    """Creates the necessary config file for auth before performing cli tests."""
    if not settings.INTEGRATION_TEST_URL:
        raise ImproperlyConfigured("INTEGRATION_TEST_URL is not set")

    host = cast(str, urlparse(settings.INTEGRATION_TEST_URL).hostname).replace(
        ".canvasmedical.com", ""
    )
    client_id = settings.INTEGRATION_TEST_CLIENT_ID
    client_secret = settings.INTEGRATION_TEST_CLIENT_SECRET

    path = CONFIG_PATH
    if not path.exists():
        if not path.parent.exists():
            path.parent.mkdir()
        path.touch()

    temp_path = path.parent / "temp_credentials.ini"

    # Read the original content
    with open(path) as original_file:
        original_content = original_file.read()

    # Write new content to the original file
    with open(path, "w") as original_file:
        original_file.writelines(
            [
                f"[{host}]\n",
                f"client_id={client_id}\n",
                f"client_secret={client_secret}\n",
            ]
        )

    # Append original content to the temp file
    with open(temp_path, "a") as temp_file:
        temp_file.write(original_content)

    yield

    with open(temp_path) as temp:
        original_content = temp.read()
        with open(path, "w") as f:
            f.write(original_content)
    os.remove(temp_path)


@pytest.fixture(autouse=True, scope="session")
def write_plugin(plugin_name: str) -> Generator[Any, Any, Any]:
    """Writes a plugin to the file system."""
    runner.invoke(app, "init", input=plugin_name)

    protocol_code = """
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol
from logger import log

class Protocol(BaseProtocol):
    RESPONDS_TO = EventType.Name(EventType.ASSESS_COMMAND__CONDITION_SELECTED)
    NARRATIVE_STRING = "I was inserted from my plugin's protocol."

    def compute(self):
        log.info(self.NARRATIVE_STRING)
        return []
"""

    with open(f"./{plugin_name}/protocols/my_protocol.py", "w") as protocol:
        protocol.write(protocol_code)

    yield

    if Path(f"./{plugin_name}").exists():
        shutil.rmtree(Path(f"./{plugin_name}"))


def list_empty_plugins(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 1 - list all plugins."""
    return ("list", 0, [], [f"{plugin_name}"])


def install_new_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 2 - install a new plugin."""
    return (
        f"install {plugin_name}",
        0,
        [
            f"Plugin {plugin_name} has a valid CANVAS_MANIFEST.json file",
            "Installing plugin:",
            "Posting",
            f"Plugin {plugin_name} successfully installed!",
        ],
        [],
    )


def list_newly_installed_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 3 - list all plugins, including newly installed one."""
    return ("list", 0, [f"{plugin_name}@0.0.1	enabled"], [])


def disable_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 4 - disable plugin."""
    return (
        f"disable {plugin_name}",
        0,
        [f"Disabling {plugin_name} using ", f"Plugin {plugin_name} successfully disabled!"],
        [],
    )


def list_disabled_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 5 - list disabled plugin."""
    return ("list", 0, [f"{plugin_name}@0.0.1	disabled"], [])


def enable_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 6 - enable the disabled plugin."""
    return (
        f"enable {plugin_name}",
        0,
        [f"Enabling {plugin_name} using ", f"Plugin {plugin_name} successfully enabled!"],
        [],
    )


def uninstall_enabled_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 7 - try to uninstall the enabled plugin."""
    return (
        f"uninstall {plugin_name}",
        1,
        [
            f"Uninstalling {plugin_name} using",
            'Status code 403: {"detail":"Cannot delete an enabled plugin."}',
        ],
        [],
    )


def uninstall_disabled_plugin(plugin_name: str) -> tuple[str, int, list[str], list[str]]:
    """Step 8 - disable and then uninstall the plugin."""
    runner.invoke(app, f"disable {plugin_name}")

    return (
        f"uninstall {plugin_name}",
        0,
        [
            f"Uninstalling {plugin_name} using",
            f"Plugin {plugin_name} successfully uninstalled!",
        ],
        [],
    )


@pytest.mark.integtest
@patch("keyring.get_password")
@patch("keyring.set_password")
@pytest.mark.parametrize(
    "step",
    [
        (list_empty_plugins),
        (install_new_plugin),
        (list_newly_installed_plugin),
        (disable_plugin),
        (list_disabled_plugin),
        (enable_plugin),
        (uninstall_enabled_plugin),
        (uninstall_disabled_plugin),
        (list_empty_plugins),
    ],
)
def test_canvas_list_install_disable_enable_uninstall(
    mock_get_password: MagicMock,
    mock_set_password: MagicMock,
    plugin_name: str,
    create_or_update_config_auth_file_for_testing: None,
    step: Callable,
) -> None:
    """Tests that the Canvas CLI can list, install, disable, enable, and uninstall a plugin."""
    mock_get_password.return_value = None
    mock_set_password.return_value = None

    (command, expected_exit_code, expected_outputs, expected_no_outputs) = step(plugin_name)

    result = runner.invoke(app, command)

    assert result.exit_code == expected_exit_code
    for expected_output in expected_outputs:
        assert expected_output in result.stdout
    for expected_no_output in expected_no_outputs:
        assert expected_no_output not in result.stdout



from ..value_set import ValueSet


class DtapVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with diphtheria, tetanus, and pertussis (DTaP) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a diphtheria, tetanus, and pertussis (DTaP) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "DTaP Vaccine"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1214"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "20",  # diphtheria, tetanus toxoids and acellular pertussis vaccine
        "50",  # DTaP-Haemophilus influenzae type b conjugate vaccine
        "106",  # diphtheria, tetanus toxoids and acellular pertussis vaccine, 5 pertussis antigens
        "107",  # diphtheria, tetanus toxoids and acellular pertussis vaccine, unspecified formulation
        "110",  # DTaP-hepatitis B and poliovirus vaccine
        "120",  # diphtheria, tetanus toxoids and acellular pertussis vaccine, Haemophilus influenzae type b conjugate, and poliovirus vaccine, inactivated (DTaP-Hib-IPV)
    }


class HepatitisAVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with Hepatitis A vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a Hepatitis A vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Hepatitis A Vaccine"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1215"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "31",  # hepatitis A vaccine, pediatric dosage, unspecified formulation
        "83",  # hepatitis A vaccine, pediatric/adolescent dosage, 2 dose schedule
        "85",  # hepatitis A vaccine, unspecified formulation
    }


class HepatitisBVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with Hepatitis B vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a Hepatitis B vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Hepatitis B Vaccine"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1216"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "08",  # hepatitis B vaccine, pediatric or pediatric/adolescent dosage
        "44",  # hepatitis B vaccine, dialysis patient dosage
        "45",  # hepatitis B vaccine, unspecified formulation
        "51",  # Haemophilus influenzae type b conjugate and Hepatitis B vaccine
        "110",  # DTaP-hepatitis B and poliovirus vaccine
    }


class HibVaccine3DoseSchedule(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with 3-dose haemophilus influenzae type b (Hib) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a 3-dose haemophilus influenzae type b (Hib) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Hib Vaccine (3 dose schedule)"
    OID = "2.16.840.1.113883.3.464.1003.110.12.1083"
    DEFINITION_VERSION = "20190315"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "49",  # Haemophilus influenzae type b vaccine, PRP-OMP conjugate
    }


class HibVaccine4DoseSchedule(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with 4-dose haemophilus influenzae type b (Hib) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a 4-dose haemophilus influenzae type b (Hib) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Hib Vaccine (4 dose schedule)"
    OID = "2.16.840.1.113883.3.464.1003.110.12.1085"
    DEFINITION_VERSION = "20190315"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "48",  # Haemophilus influenzae type b vaccine, PRP-T conjugate
        "50",  # DTaP-Haemophilus influenzae type b conjugate vaccine
        "51",  # Haemophilus influenzae type b conjugate and Hepatitis B vaccine
        "120",  # diphtheria, tetanus toxoids and acellular pertussis vaccine, Haemophilus influenzae type b conjugate, and poliovirus vaccine, inactivated (DTaP-Hib-IPV)
        "148",  # Meningococcal Groups C and Y and Haemophilus b Tetanus Toxoid Conjugate Vaccine
    }


class InactivatedPolioVaccineIpv(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with inactivated polio vaccines (IPV).

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent an inactivated polio vaccine (IPV).

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Inactivated Polio Vaccine (IPV)"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1219"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "10",  # poliovirus vaccine, inactivated
        "89",  # poliovirus vaccine, unspecified formulation
        "110",  # DTaP-hepatitis B and poliovirus vaccine
        "120",  # diphtheria, tetanus toxoids and acellular pertussis vaccine, Haemophilus influenzae type b conjugate, and poliovirus vaccine, inactivated (DTaP-Hib-IPV)
    }


class InfluenzaVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with influenza vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent an influenza vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11, CMS117v10
    """

    VALUE_SET_NAME = "Influenza Vaccine"
    OID = "2.16.840.1.113883.3.526.3.1254"
    DEFINITION_VERSION = "20170908"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "88",  # influenza virus vaccine, unspecified formulation
        "135",  # influenza, high dose seasonal, preservative-free
        "140",  # Influenza, seasonal, injectable, preservative free
        "141",  # Influenza, seasonal, injectable
        "144",  # seasonal influenza, intradermal, preservative free
        "149",  # influenza, live, intranasal, quadrivalent
        "150",  # Influenza, injectable, quadrivalent, preservative free
        "153",  # Influenza, injectable, Madin Darby Canine Kidney, preservative free
        "155",  # Seasonal, trivalent, recombinant, injectable influenza vaccine, preservative free
        "158",  # influenza, injectable, quadrivalent, contains preservative
        "161",  # Influenza, injectable,quadrivalent, preservative free, pediatric
        "166",  # influenza, intradermal, quadrivalent, preservative free, injectable
        "168",  # Seasonal trivalent influenza vaccine, adjuvanted, preservative free
        "171",  # Influenza, injectable, Madin Darby Canine Kidney, preservative free, quadrivalent
        "185",  # Seasonal, quadrivalent, recombinant, injectable influenza vaccine, preservative free
        "186",  # Influenza, injectable, Madin Darby Canine Kidney,  quadrivalent with preservative
        "197",  # influenza, high-dose seasonal, quadrivalent, .7mL dose, preservative free
        "205",  # influenza, seasonal vaccine, quadrivalent, adjuvanted, .5mL dose, preservative free
    }


class InfluenzaVirusLaivImmunization(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent live attenuated influenza vaccine.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a live attenuated influenza vaccine (quadrivalent and trivalent).

    **Exclusion Criteria:** Excludes concepts that represent the recombinant or inactivated influenza vaccine.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Influenza Virus LAIV Immunization"
    OID = "2.16.840.1.113883.3.464.1003.110.12.1087"
    DEFINITION_VERSION = "20210224"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "111",  # influenza virus vaccine, live, attenuated, for intranasal use
        "149",  # influenza, live, intranasal, quadrivalent
    }


class Measles_MumpsAndRubellaMmrVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with measles, mumps and rubella (MMR) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a measles, mumps and rubella (MMR) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Measles, Mumps and Rubella (MMR) Vaccine"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1224"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "03",  # measles, mumps and rubella virus vaccine
        "94",  # measles, mumps, rubella, and varicella virus vaccine
    }


class PneumococcalConjugateVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with pneumococcal conjugate vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a pneumococcal conjugate vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Pneumococcal Conjugate Vaccine"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1221"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "133",  # pneumococcal conjugate vaccine, 13 valent
        "152",  # Pneumococcal Conjugate, unspecified formulation
    }


class RotavirusVaccine3DoseSchedule(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with 3-dose rotavirus vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a 3-dose rotavirus vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Rotavirus Vaccine (3 dose schedule)"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1223"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "116",  # rotavirus, live, pentavalent vaccine
        "122",  # rotavirus vaccine, unspecified formulation
    }


class VaricellaZosterVaccineVzv(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with varicella zoster vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a varicella zoster vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = "Varicella Zoster Vaccine (VZV)"
    OID = "2.16.840.1.113883.3.464.1003.196.12.1170"
    DEFINITION_VERSION = "20170504"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "21",  # varicella virus vaccine
        "94",  # measles, mumps, rubella, and varicella virus vaccine
    }


class PneumococcalPolysaccharide23Vaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with pneumococcal polysaccharide vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a pneumococcal polysaccharide 23-valent vaccine.

    **Exclusion Criteria:** Excludes concepts that represent the pneumococcal conjugate 7-valent and 13-valent vaccines.

    ** Used in:** CMS127v10
    """

    VALUE_SET_NAME = "Pneumococcal Polysaccharide 23 Vaccine"
    OID = "2.16.840.1.113883.3.464.1003.110.12.1089"
    DEFINITION_VERSION = "20210224"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CVX = {
        "33",  # pneumococcal polysaccharide vaccine, 23 valent
        "109",  # pneumococcal vaccine, unspecified formulation
    }



from django.db import models

from canvas_sdk.v1.data.base import (
    CommittableModelManager,
    TimeframeLookupQuerySetMixin,
    ValueSetLookupQuerySet,
)


class TransmissionType(models.TextChoices):
    """Choices for transmission types."""

    FAX = "F", "fax"
    HL7 = "H", "hl7"
    MANUAL = "M", "manual"


class LabReport(models.Model):
    """A class representing a lab report."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_labreport_001"

    objects = CommittableModelManager()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    review_mode = models.CharField()
    junked = models.BooleanField()
    requires_signature = models.BooleanField()
    assigned_date = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="lab_reports", null=True
    )
    transmission_type = models.CharField(choices=TransmissionType)
    for_test_only = models.BooleanField()
    external_id = models.CharField()
    version = models.IntegerField()
    requisition_number = models.CharField()
    review = models.ForeignKey(
        "LabReview", on_delete=models.DO_NOTHING, related_name="reports", null=True
    )
    original_date = models.DateTimeField()
    date_performed = models.DateTimeField()
    custom_document_name = models.CharField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    deleted = models.BooleanField()


class LabReview(models.Model):
    """A class representing a lab review."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_labreview_001"

    objects = CommittableModelManager()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    internal_comment = models.TextField()
    message_to_patient = models.CharField()
    status = models.CharField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="lab_reviews", null=True
    )
    patient_communication_method = models.CharField()


class LabValueTimeframeLookupQuerySetMixin(TimeframeLookupQuerySetMixin):
    """A class that adds queryset functionality to filter using timeframes."""

    @property
    def timeframe_filter_field(self) -> str:
        """Returns the field that should be filtered on. Can be overridden for different models."""
        return "report__original_date"


class LabValueQuerySet(ValueSetLookupQuerySet, LabValueTimeframeLookupQuerySetMixin):
    """LabValueQuerySet."""

    pass


class LabValue(models.Model):
    """A class representing a lab value."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_labvalue_001"

    objects = LabValueQuerySet.as_manager()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    report = models.ForeignKey(
        "LabReport", on_delete=models.DO_NOTHING, related_name="values", null=True
    )
    value = models.TextField()
    units = models.CharField()
    abnormal_flag = models.CharField()
    reference_range = models.CharField()
    low_threshold = models.CharField()
    high_threshold = models.CharField()
    comment = models.TextField()
    observation_status = models.CharField()


class LabValueCoding(models.Model):
    """A class representing a lab value coding."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_labvaluecoding_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    value = models.ForeignKey(
        LabValue, on_delete=models.DO_NOTHING, related_name="codings", null=True
    )
    code = models.CharField()
    name = models.CharField()
    system = models.CharField()


class LabOrder(models.Model):
    """A class representing a lab order."""

    class SpecimenCollectionType(models.TextChoices):
        """Choices for specimen collection types."""

        ON_LOCATION = "L", "on location"
        PATIENT_SERVICE_CENTER = "P", "patient service center"
        OTHER = "O", "other"

    class CourtesyCopyType(models.TextChoices):
        """Choices for courtesy copy types."""

        ACCOUNT = "A", "account"
        FAX = "F", "fax"
        PATIENT = "P", "patient"

    class ManualProcessingStatus(models.TextChoices):
        """Choices for manual processing statuses."""

        MANUAL_PROCESSING_STATUS_NEEDS_REVIEW = "NEEDS_REVIEW", "Needs Review"
        MANUAL_PROCESSING_STATUS_IN_PROGRESS = "IN_PROGRESS", "In Progress"
        MANUAL_PROCESSING_STATUS_PROCESSED = "PROCESSED", "Processed"
        MANUAL_PROCESSING_STATUS_FLAGGED = "FLAGGED", "Flagged"

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_laborder_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="lab_orders", null=True
    )
    ontology_lab_partner = models.CharField()
    # TODO - uncomment when the Note model is finished
    # note = models.ForeignKey("Note", on_delete=models.DO_NOTHING, null=True)
    comment = models.CharField()
    requisition_number = models.CharField()
    is_patient_bill = models.BooleanField(null=True)
    date_ordered = models.DateTimeField()
    fasting_status = models.BooleanField(null=True)
    specimen_collection_type = models.CharField(choices=SpecimenCollectionType, null=True)
    transmission_type = models.CharField(choices=TransmissionType, null=True)
    courtesy_copy_type = models.CharField(choices=CourtesyCopyType, null=True)
    courtesy_copy_number = models.CharField()
    courtesy_copy_text = models.CharField()
    # TODO - uncomment when Staff model is added
    # ordering_provider = models.ForeignKey(Staff, on_delete=models.DO_NOTHING, null=True)
    parent_order = models.ForeignKey("LabOrder", on_delete=models.DO_NOTHING, null=True)
    healthgorilla_id = models.CharField()
    manual_processing_status = models.CharField(choices=ManualProcessingStatus)
    manual_processing_comment = models.CharField()
    labcorp_abn_url = models.URLField()


class LabOrderReason(models.Model):
    """A class representing a lab order reason."""

    class LabReasonMode(models.TextChoices):
        """Choices for lab order reasons."""

        MONITOR = "MO", "monitor"
        INVESTIGATE = "IN", "investigate"
        SCREEN_FOR = "SF", "screen for"
        UNKNOWN = "UNK", "unknown"

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_laborderreason_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    order = models.ForeignKey(
        LabOrder, on_delete=models.DO_NOTHING, related_name="reasons", null=True
    )
    mode = models.CharField(max_length=30, choices=LabReasonMode)


class LabOrderReasonCondition(models.Model):
    """A class representing a lab order reason's condition."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_laborderreasoncondition_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    reason = models.ForeignKey(
        LabOrderReason, on_delete=models.DO_NOTHING, related_name="reason_conditions", null=True
    )
    condition = models.ForeignKey(
        "v1.Condition",
        on_delete=models.DO_NOTHING,
        related_name="lab_order_reason_conditions",
        null=True,
    )


class LabTest(models.Model):
    """A class representing a lab test."""

    class LabTestOrderStatus(models.TextChoices):
        """Choicees for lab test order statuses."""

        NEW = "NE", "new"
        STAGED_FOR_REQUISITION = "SR", "staged for requisition"
        SENDING = "SE", "sending"
        SENDING_FAILED = "SF", "sending failed"
        PROCESSING = "PR", "processing"
        PROCESSING_FAILED = "PF", "processing failed"
        RECEIVED = "RE", "received"
        REVIEWED = "RV", "reviewed"
        INACTIVE = "IN", "inactive"

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_labtest_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    ontology_test_name = models.CharField(max_length=512, blank=True, default="")
    ontology_test_code = models.CharField(max_length=512, blank=True, default="")
    status = models.CharField(max_length=30, choices=LabTestOrderStatus)
    report = models.ForeignKey(
        LabReport, on_delete=models.DO_NOTHING, related_name="tests", null=True
    )
    aoe_code = models.CharField(max_length=10, default="")
    procedure_class = models.CharField(max_length=10, default="")
    specimen_type = models.CharField()
    specimen_source_code = models.CharField()
    specimen_source_description = models.CharField()
    specimen_source_coding_system = models.CharField()
    order = models.ForeignKey(
        LabOrder, on_delete=models.DO_NOTHING, related_name="tests", null=True
    )

    def __str__(self) -> str:
        return f"{self.ontology_test_name}: f{self.ontology_test_code}"



from canvas_sdk.effects import Effect, EffectType
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol


class Protocol(BaseProtocol):
    """
    You should put a helpful description of this protocol's behavior here.
    """

    # Name the event type you wish to run in response to
    RESPONDS_TO = EventType.Name(EventType.UNKNOWN)

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        from test_module_forbidden_imports_runtime_plugin.other_module.base import import_me

        return [Effect(type=EffectType.LOG, payload=import_me())]



import json
import os
import shutil
import tarfile
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, TypedDict
from urllib import parse

import psycopg
import requests
from psycopg import Connection
from psycopg.rows import dict_row

import settings
from plugin_runner.aws_headers import aws_sig_v4_headers
from plugin_runner.exceptions import InvalidPluginFormat, PluginInstallationError

# Plugin "packages" include this prefix in the database record for the plugin and the S3 bucket key.
UPLOAD_TO_PREFIX = "plugins"


def get_database_dict_from_url() -> dict[str, Any]:
    """Creates a psycopg ready dictionary from the home-app database URL."""
    parsed_url = parse.urlparse(os.getenv("DATABASE_URL"))
    db_name = parsed_url.path[1:]
    return {
        "dbname": db_name,
        "user": parsed_url.username,
        "password": parsed_url.password,
        "host": parsed_url.hostname,
        "port": parsed_url.port,
    }


def get_database_dict_from_env() -> dict[str, Any]:
    """Creates a psycopg ready dictionary from the environment variables."""
    APP_NAME = os.getenv("APP_NAME")

    return {
        "dbname": APP_NAME,
        "user": os.getenv("DB_USERNAME", "app"),
        "password": os.getenv("DB_PASSWORD", "app"),
        "host": os.getenv("DB_HOST", f"{APP_NAME}-db"),
        "port": os.getenv("DB_PORT", "5432"),
    }


def open_database_connection() -> Connection:
    """Opens a psycopg connection to the home-app database."""
    # When running within Aptible, use the database URL, otherwise pull from the environment variables.
    if os.getenv("DATABASE_URL"):
        database_dict = get_database_dict_from_url()
    else:
        database_dict = get_database_dict_from_env()
    conn = psycopg.connect(**database_dict)
    return conn


class PluginAttributes(TypedDict):
    """Attributes of a plugin."""

    version: str
    package: str
    secrets: dict[str, str]


def enabled_plugins() -> dict[str, PluginAttributes]:
    """Returns a dictionary of enabled plugins and their attributes."""
    conn = open_database_connection()

    with conn.cursor(row_factory=dict_row) as cursor:
        cursor.execute(
            "select name, package, version, key, value from plugin_io_plugin p "
            "left join plugin_io_pluginsecret s on p.id = s.plugin_id where is_enabled"
        )
        rows = cursor.fetchall()
        plugins = _extract_rows_to_dict(rows)

    return plugins


def _extract_rows_to_dict(rows: list) -> dict[str, PluginAttributes]:
    plugins = {}
    for row in rows:
        if row["name"] not in plugins:
            plugins[row["name"]] = PluginAttributes(
                version=row["version"],
                package=row["package"],
                secrets={row["key"]: row["value"]} if row["key"] else {},
            )
        else:
            plugins[row["name"]]["secrets"][row["key"]] = row["value"]
    return plugins


@contextmanager
def download_plugin(plugin_package: str) -> Generator[Path, None, None]:
    """Download the plugin package from the S3 bucket."""
    method = "GET"
    host = f"s3-{settings.AWS_REGION}.amazonaws.com"
    bucket = settings.MEDIA_S3_BUCKET_NAME
    customer_identifier = settings.CUSTOMER_IDENTIFIER
    path = f"/{bucket}/{customer_identifier}/{plugin_package}"
    payload = b"This is required for the AWS headers because it is part of the signature"
    pre_auth_headers: dict[str, str] = {}
    query: dict[str, str] = {}
    headers = aws_sig_v4_headers(
        settings.AWS_ACCESS_KEY_ID,
        settings.AWS_SECRET_ACCESS_KEY,
        pre_auth_headers,
        "s3",
        settings.AWS_REGION,
        host,
        method,
        path,
        query,
        payload,
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        prefix_dir = Path(temp_dir) / UPLOAD_TO_PREFIX
        prefix_dir.mkdir()  # create an intermediate directory reflecting the prefix
        download_path = Path(temp_dir) / plugin_package
        with open(download_path, "wb") as download_file:
            response = requests.request(method=method, url=f"https://{host}{path}", headers=headers)
            download_file.write(response.content)
        yield download_path


def install_plugin(plugin_name: str, attributes: PluginAttributes) -> None:
    """Install the given Plugin's package into the runtime."""
    try:
        print(f"Installing plugin '{plugin_name}'")

        plugin_installation_path = Path(settings.PLUGIN_DIRECTORY) / plugin_name

        # if plugin exists, first uninstall it
        if plugin_installation_path.exists():
            uninstall_plugin(plugin_name)

        with download_plugin(attributes["package"]) as plugin_file_path:
            extract_plugin(plugin_file_path, plugin_installation_path)

        install_plugin_secrets(plugin_name=plugin_name, secrets=attributes["secrets"])
    except Exception as ex:
        print(f"Failed to install plugin '{plugin_name}', version {attributes['version']}")
        raise PluginInstallationError() from ex


def extract_plugin(plugin_file_path: Path, plugin_installation_path: Path) -> None:
    """Extract plugin in `file` to the given `path`."""
    archive: tarfile.TarFile | None = None

    try:
        if tarfile.is_tarfile(plugin_file_path):
            try:
                with open(plugin_file_path, "rb") as file:
                    archive = tarfile.TarFile.open(fileobj=file)
                    archive.extractall(plugin_installation_path, filter="data")
            except tarfile.ReadError as ex:
                print(f"Unreadable tar archive: '{plugin_file_path}'")
                raise InvalidPluginFormat from ex
        else:
            print(f"Unsupported file format: '{plugin_file_path}'")
            raise InvalidPluginFormat
    finally:
        if archive:
            archive.close()


def install_plugin_secrets(plugin_name: str, secrets: dict[str, str]) -> None:
    """Write the plugin's secrets to disk in the package's directory."""
    print(f"Writing plugin secrets for '{plugin_name}'")

    secrets_path = Path(settings.PLUGIN_DIRECTORY) / plugin_name / settings.SECRETS_FILE_NAME

    # Did the plugin ship a secrets.json? TOO BAD, IT'S GONE NOW.
    if Path(secrets_path).exists():
        os.remove(secrets_path)

    with open(str(secrets_path), "w") as f:
        json.dump(secrets, f)


def disable_plugin(plugin_name: str) -> None:
    """Disable the given plugin."""
    conn = open_database_connection()
    conn.cursor().execute(
        "update plugin_io_plugin set is_enabled = false where name = %s", (plugin_name,)
    )
    conn.commit()
    conn.close()

    uninstall_plugin(plugin_name)


def uninstall_plugin(plugin_name: str) -> None:
    """Remove the plugin from the filesystem."""
    plugin_path = Path(settings.PLUGIN_DIRECTORY) / plugin_name

    if plugin_path.exists():
        shutil.rmtree(plugin_path)


def install_plugins() -> None:
    """Install all enabled plugins."""
    if Path(settings.PLUGIN_DIRECTORY).exists():
        shutil.rmtree(settings.PLUGIN_DIRECTORY)

    os.mkdir(settings.PLUGIN_DIRECTORY)

    for plugin_name, attributes in enabled_plugins().items():
        try:
            print(f"Installing plugin '{plugin_name}', version {attributes['version']}")
            install_plugin(plugin_name, attributes)
        except PluginInstallationError:
            disable_plugin(plugin_name)
            print(
                f"Installation failed for plugin '{plugin_name}', version {attributes['version']}. The plugin has been disabled"
            )
            continue

    return None



from datetime import date

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class PastSurgicalHistoryCommand(BaseCommand):
    """A class for managing a Past Surgical History command within a specific note."""

    class Meta:
        key = "surgicalHistory"
        commit_required_fields = ("past_surgical_history",)

    past_surgical_history: str | None = None
    approximate_date: date | None = None
    comment: str | None = Field(max_length=1000, default=None)

    @property
    def values(self) -> dict:
        """The Past Surgical History command's field values."""
        return {
            "past_surgical_history": self.past_surgical_history,
            "approximate_date": (
                self.approximate_date.isoformat() if self.approximate_date else None
            ),
            "comment": self.comment,
        }



from django.contrib.postgres.fields import ArrayField
from django.db import models
from timezone_utils.fields import TimeZoneField

from canvas_sdk.v1.data.common import PersonSex, TaxIDType


class Staff(models.Model):
    """Staff."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_staff_001"

    def __str__(self) -> str:
        return f"{self.first_name} {self.last_name}"

    id = models.CharField(max_length=32, db_column="key")
    dbid = models.BigIntegerField(db_column="dbid", primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    prefix = models.CharField()
    suffix = models.CharField()
    first_name = models.CharField()
    middle_name = models.CharField()
    last_name = models.CharField()
    maiden_name = models.CharField()
    nickname = models.CharField()
    previous_names = models.JSONField()
    birth_date = models.DateField(null=True)
    sex_at_birth = models.CharField(choices=PersonSex.choices)
    sexual_orientation_term = models.CharField()
    sexual_orientation_code = models.CharField()
    gender_identity_term = models.CharField()
    gender_identity_code = models.CharField()
    preferred_pronouns = models.CharField()
    biological_race_codes = ArrayField(models.CharField())
    biological_race_terms = ArrayField(models.CharField())
    cultural_ethnicity_codes = ArrayField(models.CharField())
    cultural_ethnicity_terms = ArrayField(models.CharField())
    last_known_timezone = TimeZoneField(null=True)
    active = models.BooleanField()
    # TODO - uncomment when PracticeLocation field is developed
    # primary_practice_location = models.ForeignKey(
    #     'v1.PracticeLocation', on_delete=models.DO_NOTHING, null=True
    # )
    npi_number = models.CharField()
    nadean_number = models.CharField()
    group_npi_number = models.CharField()
    bill_through_organization = models.BooleanField()
    tax_id = models.CharField()
    tax_id_type = models.CharField(choices=TaxIDType.choices)
    spi_number = models.CharField()
    # TODO - uncomment when Language is developed
    # language = models.ForeignKey('v1.Language', on_delete=models.DO_NOTHING, related_name="staff_speakers", null=True)
    # language_secondary = models.ForeignKey('v1.Language', on_delete=models.DO_NOTHING, related_name="staff_secondary_speakers", null=True)
    personal_meeting_room_link = models.URLField(null=True)
    state = models.JSONField()
    user = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    schedule_column_ordering = models.IntegerField()
    default_supervising_provider = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="supervising_team", null=True
    )



from ..value_set import ValueSet


class Ethnicity(ValueSet):
    """Ethnicity value set.

    **Clinical Focus:**

    **Data Element Scope:**

    **Inclusion Criteria:**

    **Exclusion Criteria:**

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = "Ethnicity"
    OID = "2.16.840.1.114222.4.11.837"
    DEFINITION_VERSION = "20121025"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CDCREC = {
        "2135-2",  # Hispanic or Latino
        "2186-5",  # Not Hispanic or Latino
    }


class OncAdministrativeSex(ValueSet):
    """ONC Administrative Sex value set.

    **Clinical Focus:** Gender identity restricted to only Male and Female used in administrative situations requiring a restriction to these two categories.

    **Data Element Scope:** Gender

    **Inclusion Criteria:** Male and Female only.

    **Exclusion Criteria:** Any gender identity that is not male or female.

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = "ONC Administrative Sex"
    OID = "2.16.840.1.113762.1.4.1"
    DEFINITION_VERSION = "20150331"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    ADMINISTRATIVEGENDER = {
        "F",  # Female
        "M",  # Male
    }


class Payer(ValueSet):
    """Payer value set.

    **Clinical Focus:** Categories of types of health care payor entities as defined by the US Public Health Data Consortium SOP code system

    **Data Element Scope:** @code in CCDA r2.1 template Planned Coverage [act: identifier urn:oid:2.16.840.1.113883.10.20.22.4.129 (open)] DYNAMIC

    **Inclusion Criteria:** All codes in the code system

    **Exclusion Criteria:** none

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = "Payer"
    OID = "2.16.840.1.114222.4.11.3591"
    DEFINITION_VERSION = "20180718"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    SOP = {
        "1",  # MEDICARE
        "2",  # MEDICAID
        "3",  # OTHER GOVERNMENT (Federal/State/Local) (excluding Department of Corrections)
        "4",  # DEPARTMENTS OF CORRECTIONS
        "5",  # PRIVATE HEALTH INSURANCE
        "6",  # BLUE CROSS/BLUE SHIELD
        "7",  # MANAGED CARE, UNSPECIFIED (to be used only if one can't distinguish public from private)
        "8",  # NO PAYMENT from an Organization/Agency/Program/Private Payer Listed
        "9",  # MISCELLANEOUS/OTHER
        "11",  # Medicare Managed Care (Includes Medicare Advantage Plans)
        "12",  # Medicare (Non-managed Care)
        "13",  # Medicare Hospice
        "14",  # Dual Eligibility Medicare/Medicaid Organization
        "19",  # Medicare Other
        "21",  # Medicaid (Managed Care)
        "22",  # Medicaid (Non-managed Care Plan)
        "23",  # Medicaid/SCHIP
        "25",  # Medicaid - Out of State
        "26",  # Medicaid - Long Term Care
        "29",  # Medicaid Other
        "31",  # Department of Defense
        "32",  # Department of Veterans Affairs
        "33",  # Indian Health Service or Tribe
        "34",  # HRSA Program
        "35",  # Black Lung
        "36",  # State Government
        "37",  # Local Government
        "38",  # Other Government (Federal, State, Local not specified)
        "39",  # Other Federal
        "41",  # Corrections Federal
        "42",  # Corrections State
        "43",  # Corrections Local
        "44",  # Corrections Unknown Level
        "51",  # Managed Care (Private)
        "52",  # Private Health Insurance - Indemnity
        "53",  # Managed Care (private) or private health insurance (indemnity), not otherwise specified
        "54",  # Organized Delivery System
        "55",  # Small Employer Purchasing Group
        "56",  # Specialized Stand-Alone Plan
        "59",  # Other Private Insurance
        "61",  # BC Managed Care
        "62",  # BC Insurance Indemnity
        "71",  # HMO
        "72",  # PPO
        "73",  # POS
        "79",  # Other Managed Care
        "81",  # Self-pay (Includes applicants for insurance and Medicaid applicants)
        "82",  # No Charge
        "83",  # Refusal to Pay/Bad Debt
        "84",  # Hill Burton Free Care
        "85",  # Research/Donor
        "89",  # No Payment, Other
        "91",  # Foreign National
        "92",  # Other (Non-government)
        "93",  # Disability Insurance
        "94",  # Long-term Care Insurance
        "95",  # Worker's Compensation
        "96",  # Auto Insurance (includes no fault)
        "97",  # Legal Liability / Liability Insurance
        "98",  # Other specified but not otherwise classifiable (includes Hospice - Unspecified plan)
        "99",  # No Typology Code available for payment source
        "111",  # Medicare HMO
        "112",  # Medicare PPO
        "113",  # Medicare POS
        "119",  # Medicare Managed Care Other
        "121",  # Medicare FFS
        "122",  # Medicare Drug Benefit
        "123",  # Medicare Medical Savings Account (MSA)
        "129",  # Medicare Non-managed Care Other
        "191",  # Medicare Pharmacy Benefit Manager
        "211",  # Medicaid HMO
        "212",  # Medicaid PPO
        "213",  # Medicaid PCCM (Primary Care Case Management)
        "219",  # Medicaid Managed Care Other
        "291",  # Medicaid Pharmacy Benefit Manager
        "299",  # Medicaid - Dental
        "311",  # TRICARE (CHAMPUS)
        "312",  # Military Treatment Facility
        "313",  # Dental --Stand Alone
        "321",  # Veteran care-Care provided to Veterans
        "322",  # Non-veteran care
        "331",  # Indian Health Service - Regular
        "332",  # Indian Health Service - Contract
        "333",  # Indian Health Service - Managed Care
        "334",  # Indian Tribe - Sponsored Coverage
        "341",  # Title V (MCH Block Grant)
        "342",  # Migrant Health Program
        "343",  # Ryan White Act
        "349",  # Other
        "361",  # State SCHIP program (codes for individual states)
        "362",  # Specific state programs (list/ local code)
        "369",  # State, not otherwise specified (other state)
        "371",  # Local - Managed care
        "372",  # FFS/Indemnity
        "379",  # Local, not otherwise specified (other local, county)
        "381",  # Federal, State, Local not specified managed care
        "382",  # Federal, State, Local not specified - FFS
        "389",  # Federal, State, Local not specified - Other
        "391",  # Federal Employee Health Plan - Use when known
        "511",  # Commercial Managed Care - HMO
        "512",  # Commercial Managed Care - PPO
        "513",  # Commercial Managed Care - POS
        "514",  # Exclusive Provider Organization
        "515",  # Gatekeeper PPO (GPPO)
        "516",  # Commercial Managed Care - Pharmacy Benefit Manager
        "517",  # Commercial Managed Care - Dental
        "519",  # Managed Care, Other (non HMO)
        "521",  # Commercial Indemnity
        "522",  # Self-insured (ERISA) Administrative Services Only (ASO) plan
        "523",  # Medicare supplemental policy (as second payer)
        "524",  # Indemnity Insurance - Dental
        "529",  # Private health insurance--other commercial Indemnity
        "561",  # Dental
        "562",  # Vision
        "611",  # BC Managed Care - HMO
        "612",  # BC Managed Care - PPO
        "613",  # BC Managed Care - POS
        "614",  # BC Managed Care - Dental
        "619",  # BC Managed Care - Other
        "621",  # BC Indemnity
        "622",  # BC Self-insured (ERISA) Administrative Services Only (ASO)Plan
        "623",  # BC Medicare Supplemental Plan
        "629",  # BC Indemnity - Dental
        "821",  # Charity
        "822",  # Professional Courtesy
        "823",  # Research/Clinical Trial
        "951",  # Worker's Comp HMO
        "953",  # Worker's Comp Fee-for-Service
        "954",  # Worker's Comp Other Managed Care
        "959",  # Worker's Comp, Other unspecified
        "3111",  # TRICARE Prime--HMO
        "3112",  # TRICARE Extra--PPO
        "3113",  # TRICARE Standard - Fee For Service
        "3114",  # TRICARE For Life--Medicare Supplement
        "3115",  # TRICARE Reserve Select
        "3116",  # Uniformed Services Family Health Plan (USFHP) -- HMO
        "3119",  # Department of Defense - (other)
        "3121",  # Enrolled Prime--HMO
        "3122",  # Non-enrolled Space Available
        "3123",  # TRICARE For Life (TFL)
        "3211",  # Direct Care-Care provided in VA facilities
        "3212",  # Indirect Care-Care provided outside VA facilities
        "3221",  # Civilian Health and Medical Program for the VA (CHAMPVA)
        "3222",  # Spina Bifida Health Care Program (SB)
        "3223",  # Children of Women Vietnam Veterans (CWVV)
        "3229",  # Other non-veteran care
        "3711",  # HMO
        "3712",  # PPO
        "3713",  # POS
        "3811",  # Federal, State, Local not specified - HMO
        "3812",  # Federal, State, Local not specified - PPO
        "3813",  # Federal, State, Local not specified - POS
        "3819",  # Federal, State, Local not specified - not specified managed care
        "9999",  # Unavailable / No Payer Specified / Blank
        "32121",  # Fee Basis
        "32122",  # Foreign Fee/Foreign Medical Program (FMP)
        "32123",  # Contract Nursing Home/Community Nursing Home
        "32124",  # State Veterans Home
        "32125",  # Sharing Agreements
        "32126",  # Other Federal Agency
        "32127",  # Dental Care
        "32128",  # Vision Care
    }


class Race(ValueSet):
    """Race value set.

    **Clinical Focus:**

    **Data Element Scope:**

    **Inclusion Criteria:**

    **Exclusion Criteria:**

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = "Race"
    OID = "2.16.840.1.114222.4.11.836"
    DEFINITION_VERSION = "20121025"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CDCREC = {
        "1002-5",  # American Indian or Alaska Native
        "2028-9",  # Asian
        "2054-5",  # Black or African American
        "2076-8",  # Native Hawaiian or Other Pacific Islander
        "2106-3",  # White
        "2131-1",  # Other Race
    }


class Female(ValueSet):
    """Female Gender value set.

    **Clinical Focus:** Concepts that represent Female when assessing quality measures

    **Data Element Scope:** Gender

    **Inclusion Criteria:** Appropriate female gender concepts

    **Exclusion Criteria:** Concepts representing Male gender

    ** Used in:** CMS124v10, CMS153v10, CMS125v10, CMS249v4
    """

    VALUE_SET_NAME = "Female"
    OID = "2.16.840.1.113883.3.560.100.2"
    DEFINITION_VERSION = "20160331"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    ADMINISTRATIVEGENDER = {
        "F",  # Female
    }


class White(ValueSet):
    """White Race value set.

    **Clinical Focus:** The purpose of this value set is to represent concepts for the patient characteristic of white race.

    **Data Element Scope:** This value set may use a model element related to Race.

    **Inclusion Criteria:** Includes concepts that represent a patient characteristic of white race.

    **Exclusion Criteria:** Excludes concepts that represent non-white race.

    ** Used in:** CMS249v4
    """

    VALUE_SET_NAME = "White"
    OID = "2.16.840.1.113883.3.464.1003.123.12.1007"
    DEFINITION_VERSION = "20180321"
    EXPANSION_VERSION = "eCQM Update 2021-05-06"

    CDCREC = {
        "2106-3",  # White
    }



from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class PerformCommand(BaseCommand):
    """A class for managing a Perform command within a specific note."""

    class Meta:
        key = "perform"
        commit_required_fields = ("cpt_code",)

    cpt_code: str
    notes: str | None = None

    @property
    def values(self) -> dict:
        """The Perform command's field values."""
        return {"cpt_code": self.cpt_code, "notes": self.notes}



import json
import tarfile
import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

from pytest_mock import MockerFixture

from plugin_runner.plugin_installer import (
    PluginAttributes,
    _extract_rows_to_dict,
    download_plugin,
    install_plugins,
    uninstall_plugin,
)


def _create_tarball(name: str) -> Path:
    # Create a temporary tarball file
    temp_dir = tempfile.mkdtemp()
    tarball_path = Path(temp_dir) / f"{name}.tar.gz"

    # Add some files to the tarball
    with tarfile.open(tarball_path, "w:gz") as tar:
        for i in range(3):
            file_path = Path(temp_dir) / f"file{i}.txt"
            file_path.write_text(f"Content of file {i}")
            tar.add(file_path, arcname=f"file{i}.txt")

    # Return a Path handle to the tarball
    return tarball_path


def test_extract_rows_to_dict() -> None:
    """Test that database rows can be extracted to a dictionary with secrets appropriately attributed to plugin."""
    rows = [
        {
            "name": "plugin1",
            "version": "1.0",
            "package": "package1",
            "key": "key1",
            "value": "value1",
        },
        {
            "name": "plugin1",
            "version": "1.0",
            "package": "package1",
            "key": "key2",
            "value": "value2",
        },
        {"name": "plugin2", "version": "2.0", "package": "package2", "key": None, "value": None},
    ]

    expected_output = {
        "plugin1": {
            "version": "1.0",
            "package": "package1",
            "secrets": {"key1": "value1", "key2": "value2"},
        },
        "plugin2": {
            "version": "2.0",
            "package": "package2",
            "secrets": {},
        },
    }

    result = _extract_rows_to_dict(rows)
    assert result == expected_output


def test_plugin_installation_from_tarball(mocker: MockerFixture) -> None:
    """Test that plugins can be installed from tarballs."""
    mock_plugins = {
        "plugin1": PluginAttributes(
            version="1.0", package="plugins/plugin1.tar.gz", secrets={"key1": "value1"}
        ),
        "plugin2": PluginAttributes(
            version="1.0", package="plugins/plugin2.tar", secrets={"key2": "value2"}
        ),
    }

    tarball_1 = _create_tarball("plugin1")
    tarball_2 = _create_tarball("plugin2")

    mocker.patch("plugin_runner.plugin_installer.enabled_plugins", return_value=mock_plugins)

    def mock_download_plugin(package: str) -> MagicMock:
        mock_context = mocker.Mock()
        if package == "plugins/plugin1.tar.gz":
            mock_context.__enter__ = mocker.Mock(return_value=tarball_1)
        elif package == "plugins/plugin2.tar":
            mock_context.__enter__ = mocker.Mock(return_value=tarball_2)
        mock_context.__exit__ = mocker.Mock(return_value=None)
        return mock_context

    mocker.patch(
        "plugin_runner.plugin_installer.download_plugin",
        side_effect=mock_download_plugin,
    )

    install_plugins()
    assert Path("plugin_runner/tests/data/plugins/plugin1").exists()
    assert Path("plugin_runner/tests/data/plugins/plugin1/SECRETS.json").exists()
    with open("plugin_runner/tests/data/plugins/plugin1/SECRETS.json") as f:
        assert json.load(f) == mock_plugins["plugin1"]["secrets"]
    assert Path("plugin_runner/tests/data/plugins/plugin2").exists()
    assert Path("plugin_runner/tests/data/plugins/plugin2/SECRETS.json").exists()
    with open("plugin_runner/tests/data/plugins/plugin2/SECRETS.json") as f:
        assert json.load(f) == mock_plugins["plugin2"]["secrets"]

    uninstall_plugin("plugin1")
    uninstall_plugin("plugin2")
    assert not Path("plugin_runner/tests/data/plugins/plugin1").exists()
    assert not Path("plugin_runner/tests/data/plugins/plugin2").exists()


def test_download(mocker: MockerFixture) -> None:
    """Test that the plugin package can be written to disk, mocking out S3."""
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.content = b"some content in a file"
    with patch("requests.request", return_value=mock_response) as mock_request:
        plugin_package = "plugins/plugin1.tar.gz"
        with download_plugin(plugin_package) as plugin_path:
            assert plugin_path.exists()
            assert plugin_path.read_bytes() == b"some content in a file"
        mock_request.assert_called_once()
